/*
 * logrunner.c
 * (c) Peter Roozemaal, feb 2003
 *
 * $Id: logrunner.cpp,v 1.6 2007-12-10 16:12:37 arjen Exp $
 *
 * 1) compile,
 * 2) Add 'logfile' elements to the gnucomo configuration file
 *    defining the name and type of each logfile to scan.
 *    'filter' elements can be addes to pre-filter lines from the log.
 *
 *    example:
 *       <logfile>
 *          <name>/var/log/httpd/access_log</name>
 *          <type>apache access log</type>
 *       </logfile>
 *       <logfile>
 *          <name>/var/log/messages</name>
 *          <type>system log</type>
 *          <filter>open_scanner</filter>
 *          <filter>session closed</filter>
 *       </logfile>
 *
 * 3) run executable and feed the output to gcm_input
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>

#include <fstream>
#include <list>

#include <AXE/String.h>

#include "gnucomo_config.h"

extern String XML_Entities(String s);

static const char* const usage =
   "Usage: logrunner [<options>] [<hostname>]\n\n"
   "<hostname> is the gnucomo data collection server\n"
   "Options are:\n"
   " -1: one-shot; stop after a single pass over one of the configured logfiles\n"
   " -c <name>: specify alternative configuration name, default is \'gnucomo\'\n"
   " -l <file>: Read input from logfile <file>.\n"
   " -p <number>: set IP port number to connect to\n"
   " -s <file>: Use <file> as the status file instead of the default.\n"
   " -v Verbose output\n"
   "If no hostname is specified, logrunner sends output to stdout\n";

class LogFile {
   String   name;
   String   type;
   String   fromhost;

   int     fd;
   ino_t   inode;
   off_t   position;

   std::list<regex> filter;

   void copy_data();

public:
   LogFile(String n, String t, String fh)
   {
      name = n;
      type = t;
      fromhost = fh;
      inode     = 0;
      position  = 0;
      fd        = -1;
   }

   LogFile(const LogFile &lf)
   {
      name     = lf.name;
      type     = lf.type;
      fromhost = lf.fromhost;

      fd       = lf.fd;
      inode    = lf.inode;
      position = lf.position;
      filter   = lf.filter;
   }

   String pathname()
   {
      return name;
   }

   void update_status(ino_t i, off_t p)
   {
      inode     = i;
      position  = p;
   }

   String status()
   {
      return name + " " + String(inode) + " " + String(position);
   }

   void add_filter(String filter_expression)
   {
      filter.push_back(filter_expression);
   }

   void do_file();
};

static std::list<LogFile>     logs_to_run;

String confname = "gnucomo";
static char* statusname = "/var/lib/logrunner.status";
static char* newstatusname = "/var/lib/logrunner.status.new";

static const char *hostname = NULL;
static int port = 2996;         /* random magic number */
static int out_stream = 1;

static int oneshot = 0;
static int verbose = 0;

static int something_has_changed = 0;
   
static volatile sig_atomic_t sig_seen = 0;

static void sighandler(int sig)
{
   sig_seen = sig;
}

static void set_signal_handler(void)
{
   /* These aren't all; but it's a nice set to start with */
   signal(SIGHUP, sighandler);
   signal(SIGINT, sighandler);
   signal(SIGQUIT, sighandler);
   signal(SIGABRT, sighandler);
   signal(SIGPIPE, sighandler);
   signal(SIGTERM, sighandler);
}


void open_output()
{
   struct hostent* hostptr;
   struct sockaddr_in addr;
   unsigned int namelen = sizeof(addr);

   if ( hostname )
   {
      hostptr = gethostbyname(hostname);
      if ( !hostptr )
      {
         fprintf(stderr, "logrunner: FATAL: cannot resolve %s\n", hostname);
         exit(2);
      }
      out_stream = socket(PF_INET, SOCK_STREAM, 0);
      if ( out_stream < 0 )
      {
         fprintf(stderr, "logrunner: FATAL: Socket creation failed\n");
         exit(2);
      }
      addr.sin_family = AF_INET;
      addr.sin_addr.s_addr = *((long*)(hostptr->h_addr));
      addr.sin_port = htons(port);
      if ( connect(out_stream, (struct sockaddr*) &addr, namelen) < 0 )
      {
         fprintf(stderr, "logrunner: FATAL: connect to %s failed\n", hostname);
         exit(2);
      }
   }
   else
   {
      out_stream = 1;
   }
}

void xsend(const char* str)
{
   write(out_stream, str, strlen(str));
}

void xml_header(String type)
{
   struct hostent *host;
   char buffer[256];
   *buffer = 0;
   gethostname(buffer, sizeof(buffer));

   //  Try to obtain the official name of the host (FQDN)
 
   host = gethostbyname(buffer);
   if (host != NULL)
   {
      strcpy(buffer, host->h_name);
   }

   xsend("<?xml version='1.0'?>\n");
   xsend("<gcmt:message xmlns:gcmt=\"http://gnucomo.org/transport/\">\n");
   xsend("<gcmt:header>\n<gcmt:hostname>");
   xsend(buffer);
   xsend("</gcmt:hostname>\n<gcmt:messagetype>");
   xsend(type);
   xsend("</gcmt:messagetype>\n</gcmt:header>\n");
   xsend("<gcmt:data><gcmt:log>");
}

void xml_footer(void)
{
   xsend("</gcmt:log></gcmt:data>\n</gcmt:message>\n");
}

void output(const char* str)
{
   std::cerr << str;
}

void output_error(const char* str)
{
   std::cerr << str << " errno=" << errno << "\n";
}

void LogFile::do_file()
{
   struct stat statinfo;
   char buffer[80];

   /* inode check */
   if ( stat(name, &statinfo) )
   {
      std::cerr << "!!! logfile: stat failed: " << name << ", errno=" << errno << "\n";
   }
   else
   {
      if ( statinfo.st_ino != inode )
      {
         if ( fd >= 0 )
         {
            copy_data();
            close(fd);
            fd = -1;
         }
         if (verbose)
         {
            std::cerr << "@@@ logfile: logrotate detected: " << name << "\n";
         }
         inode = statinfo.st_ino;
         position = 0;
      }
   }
   if ( fd < 0 )
   {
      /* attempt to open the file */
      fd = open(name, O_RDONLY);
      if ( fd < 0 )
      {
         std::cerr << "!!! logfile: open failed: " << name << ", " << strerror(errno) << "\n";
         return;
      }
      lseek(fd, position, SEEK_SET);

      if (verbose)
      {
         std::cerr << "*** logfile: opened: " << name;
         std::cerr << "\n*** logfile: resumed read from position ";
         std::cerr << position << "\n";
         std::cerr << "This logfile has " << filter.size() << " filters.\n";
      }
   }

   copy_data();
}

void LogFile::copy_data()
{
   char buffer[4096];
   int ndata;
   String logline("");

   /* read data and dump to output */
   ndata = read(fd, buffer, sizeof(buffer)-1);
   if ( ndata > 0 )
   {
      xml_header(type);
      while ( ndata > 0 )
      {
         //  Make a separate <gcmt:raw> element from each line
 
         char *line, *nextline;

         line = buffer;
         nextline = buffer;

         while (nextline < buffer + ndata)
         {
            while (*nextline != '\n' && nextline < buffer + ndata)
            {
               nextline++;
            }
            if (*nextline == '\n')
            {
               // Another line found - make the split.
               *nextline++ = '\0';

               logline += line;

               // See if we have to select the host and apply filters to this log entry

               bool filtered_out = false;

               if (fromhost && (fromhost.in(logline) == -1 || fromhost.in(logline) > 20))
               {
                  filtered_out = true;
               }

               std::list<regex>::iterator f = filter.begin();
               while (f != filter.end())
               {
                  if (logline == *f)
                  {
                     filtered_out = true;
                  }
                  f++;
               }

               if (!filtered_out)
               {
                  logline = XML_Entities(logline);

                  write(out_stream, "<gcmt:raw>", 10);
                  write(out_stream, logline, ~logline);
                  write(out_stream, "</gcmt:raw>\n", 12);
               }
               else if (verbose)
               {
                  std::cerr << logline << " is filtered out.\n";
               }

               line = nextline;
               logline = "";
            }
            else
            {
               //  We have a buffer full of data but no newline.
               //  Flush the buffer into logline and continue reading.

               nextline[0] = '\0';
               logline += line;
               line = nextline;
            }
         }
         if (line != nextline)
         {
            //  There is still an incomplete line in the buffer.
            memmove(buffer, line, nextline - line);
         }
         position += ndata - (nextline - line);
         ndata -= line - buffer;
         ndata += read(fd, buffer + (nextline - line), sizeof(buffer) - 1 - (nextline - line));
      }
      xml_footer();
      something_has_changed = 1;
   }   
   if ( ndata < 0 )
   {
      std::cerr << "!!! logfile: read failed: " << name << ", " << strerror(errno) << "\n";
      std::cerr << "    file descriptor = " << fd << "\n";
   }
}

void write_status_file()
{
   FILE* dumpfile;
   int localerror = 0;

   std::ofstream   statusfile(newstatusname);
   std::list<LogFile>::iterator lf = logs_to_run.begin();
   while (lf != logs_to_run.end())
   {
      if (verbose)
      {
         std::cerr  << "Write status for " << lf->pathname() << "\n";
      }
      statusfile << lf->status() << "\n";
      lf++;
   }
   
   if ( localerror == 0 )
   {
      if ( rename(newstatusname, statusname) )
      {
         output_error("!!! dumpstatus: rename failed");
      }
   }
   something_has_changed = 0;
}

void read_status()
{
   FILE* statusfile;
   char buffer[4096];
   char* xp;
   long ino;
   unsigned long pos;
   struct fileinfo* fp;

   statusfile = fopen(statusname, "r");
   if ( statusfile == NULL )
   {
      fprintf(stderr, "logrunner: can\'t open status file \'%s\': ",
         statusname);
      perror("");
      return;
   }
   while ( fgets(buffer, sizeof(buffer), statusfile ) )
   {
      xp = strchr(buffer, ' ');
      if ( xp )
      {
         *xp = 0;
         sscanf(xp+1, "%ld %lu", &ino, &pos);

         //  Search for the logfile in the list of logs to run

         std::list<LogFile>::iterator lf = logs_to_run.begin();
         while (lf != logs_to_run.end())
         {
            if (lf->pathname() == String(buffer))
            {
               if (verbose)
               {
                  std::cerr << "Read status for " << lf->pathname() << "\n";
               }
               lf->update_status(ino, pos);
            }
            lf++;
         }

      }
   }
   fclose(statusfile);
}

void read_config(gnucomo_config cfg)
{
   /*
    *   The configuration for logrunner is stored in the central Gnucomo
    *   configuration file. Multiple 'logfile' elements can be put in this
    *   XML file, one for each logfile to scan.
    *   Each 'logfile' element has at least a 'name' and a 'type' element
    *   that denote the pathname of the logfile and the type of its content.
    */

   String logfilename;
   String logfiletype;
   String fromhost;

   int l = 0;

   logfilename = cfg.find_parameter("logfile", "name", l);
   while (logfilename != String(""))
   {
      logfiletype = cfg.find_parameter("logfile", "type", l);
      fromhost = cfg.find_parameter("logfile", "fromhost", l);
      if (verbose)
      {
         std::cerr << "LogFile " << logfilename << " of type "
                   << logfiletype << " from host " << fromhost << "\n";
      }

      LogFile lf(logfilename, logfiletype, fromhost);

      int f = 0;
      String exp  = cfg.find_parameter("logfile", "filter", l, f);
      while (exp != "")
      {
         lf.add_filter(exp);

         f++;
         exp = cfg.find_parameter("logfile", "filter", l, f);
      }

      logs_to_run.push_back(lf);

      l++;
      logfilename = cfg.find_parameter("logfile", "name", l);
   }

}

void process_options(int argc, char* argv[])
{
   const char* const options = "1c:l:p:s:v";
   int opt;

   String logfilename;

   opt = getopt(argc, argv, options);
   while ( opt != -1 )
   {
      switch ( opt )
      {
      case '1':
         oneshot = 1;
         break;
      case 'c':
         confname = strdup(optarg);
         break;
      case 'l':
         logfilename = optarg;
         break;
      case 'p':
         port = atoi(optarg);
         break;
      case 's':
         statusname = strdup(optarg);
         newstatusname = (char *)malloc(strlen(statusname) + 6);
         strcpy(newstatusname, statusname);
         strcat(newstatusname, ".new");
         break;
      case 'v':
         verbose = 1;
         break;
      default:
         fputs(usage, stderr);
         exit(2);
      }
      opt = getopt(argc, argv, options);
   }
   switch ( argc-optind )
   {
   case 0:
      break;
   case 1:
      hostname = argv[optind];
      break;
   default:
      fputs(usage, stderr);
      exit(2);
   }

   if (logfilename)
   {
      LogFile lf(logfilename, String("system log"), String(""));
      logs_to_run.push_back(lf);
   }
}

int main(int argc, char* argv[])
{

   gnucomo_config    cfg;

   process_options(argc, argv);
   open_output();
   
   if (!cfg.read(confname))
   {
      std::cerr << "Can not read Gnucomo configuration file for " << confname << ".\n";
      exit(1);
   }

   if (logs_to_run.empty())
   {
      read_config(cfg);
   }
   read_status();

   set_signal_handler();

   while ( sig_seen == 0 )
   {
      std::list<LogFile>::iterator lf = logs_to_run.begin();
      while (lf != logs_to_run.end() && !(something_has_changed && oneshot))
      {
         lf->do_file();
         lf++;
      }

      if ( something_has_changed )
      {
         write_status_file();
      }
      if ( oneshot )
      {
         return 0;
      }
      sleep(1);
   }

   fprintf(stderr, "logrunner: stopped by signal %d\n", sig_seen);
   /* shouldn't we close files and release memory here? */
   return 0;
}
