
/**************************************************************************
**  (c) Copyright 2003, Andromeda Technology & Automation
** This is free software; you can redistribute it and/or modify it under the
** terms of the GNU General Public License, see the file COPYING.
***************************************************************************
** MODULE INFORMATION *
***********************
**      FILE NAME      : rpm_filter.cpp
**      SYSTEM NAME    : 
**      VERSION NUMBER : $Revision: 1.2 $
**
**  DESCRIPTION      :  Transform a list of packages into a Gnucomo XML document
**
**  EXPORTED OBJECTS : 
**  LOCAL    OBJECTS : 
**  MODULES  USED    :
***************************************************************************
**  ADMINISTRATIVE INFORMATION *
********************************
**      ORIGINAL AUTHOR : Arjen Baart - arjen@andromeda.nl
**      CREATION DATE   : Nov 27, 2003
**      LAST UPDATE     : Nov 12, 2007
**      MODIFICATIONS   : 
**************************************************************************/

/*****************************
   $Log: rpm_filter.cpp,v $
   Revision 1.2  2007-11-14 16:19:25  arjen
   Bugfix: Segmentation fault when reading an rpm package list
   with empty lines.

   Revision 1.1  2003/12/04 10:38:09  arjen
   Major redesign. All input is handled through XML. Raw input data is first
   transformed into an XML document for further processing.
   A collection of polymorphic classes handle the transformation of various
   input formats into XML.
   Classifying input data is done with a finite improbability calculation.

*****************************/

/* static const char *RCSID = "$Id: rpm_filter.cpp,v 1.2 2007-11-14 16:19:25 arjen Exp $"; */

#include <ctype.h>

#include "rpm_filter.h"

/*=========================================================================
**  NAME           : constructXML
**  SYNOPSIS       : int constructXML(message_buffer &in, strstream &xml)
**  PARAMETERS     : 
**  RETURN VALUE   : Create an XML document from a list of RPMs
**
**  DESCRIPTION    : 
**                   Scan a list of packages and versions from "rpm -a".
**                   A similar listing can be created on IRIX 6.5 by using the
**                   command "showprods -3 -n|awk '{printf "%s-%s\n",$2,$3}'|grep -v '^[-=]' \
**                             |grep -v Version-Description".
**                 
**                   We have to separate the package name and the version.
**                   The separation is marked by a '-', followed by a digit.
**                   However, there may be other sequences of '-'digit in the package name,
**                   do we have to scan ahead until there is at most one such sequence
**                   left in the version string. The '-'digit seqeunce inside the
**                   version usually separates the version and the release number.
**
**  VARS USED      :
**  VARS CHANGED   :
**  FUNCTIONS USED :
**  SEE ALSO       :
**  LAST MODIFIED  : Nov 12, 2007
**=========================================================================
*/

static const regex re_rpm("[[:alnum:]+-]+-[0-9][[:alnum:].-]");

void rpm_filter::construct_XML(message_buffer &in, std::strstream &xml)
{
   String line;

   scan_email_header(in);
   construct_header(xml);

   xml << "  <gcmt:data>\n";
   xml << "    <gcmt:parameters class='package'>\n";

   while (in >> line)
   { 
      int  version_start, next_version_start;
      int  i;

      if (line == re_rpm)
      {
         //   Separate the package name from the version number
 
         i = line.index('-');

         version_start = i;
         next_version_start = i;

         while (i < ~line - 1)
         {
            while (i < ~line - 1 && !(line[i] == '-' && isdigit(line[i + 1])))
            {
               i++;
            }
            if (i < ~line - 1)
            {
               version_start = next_version_start;
               next_version_start = i;
            }
            i++;
         }
      
         if (!isdigit(line[version_start + 1]))
         {
            version_start = next_version_start;
         }
         String package(line(0,version_start));
         String version(line(version_start + 1, ~line));

         //   Create the XML element.

         xml << "    <gcmt:parameter name='" << package << "'>\n";
         xml << "      <gcmt:property name='version'>" << version << "</gcmt:property>\n";
         xml << "    </gcmt:parameter>\n";
      }
   }
   xml << "    </gcmt:parameters>\n";
   xml << "  </gcmt:data>\n";
   xml << "</gcmt:message>\n";
}

bool rpm_cooker::check_pattern(String logline)
{
   return logline == re_rpm;
}

bool rpm_cooker::cook_this(String logline, UTC arrival)
{
   return true;
}
