
/**************************************************************************
**  (c) Copyright 2002, Andromeda Technology & Automation
** This is free software; you can redistribute it and/or modify it under the
** terms of the GNU General Public License, see the file COPYING.
***************************************************************************
** MODULE INFORMATION *
***********************
**      FILE NAME      : database.h
**      SYSTEM NAME    : 
**      VERSION NUMBER : $Revision: 1.12 $
**
**  DESCRIPTION      :  Classes to provide an abstract layer on the Gnucomo
**                      database.
**
**  EXPORTED OBJECTS : 
**  LOCAL    OBJECTS : 
**  MODULES  USED    :
***************************************************************************
**  ADMINISTRATIVE INFORMATION *
********************************
**      ORIGINAL AUTHOR : Arjen Baart - arjen@andromeda.nl
**      CREATION DATE   : Sep 10, 2002
**      LAST UPDATE     : Aug 19, 2005
**      MODIFICATIONS   : 
**************************************************************************/

/*****************************
   $Log: database.h,v $
   Revision 1.12  2011-03-24 10:21:43  arjen
   Adjusted to new version of libpqxx.

   Revision 1.11  2007/01/11 13:50:10  arjen
   Experimental start of database OO abstraction layer.

   Revision 1.10  2003/12/04 10:39:36  arjen
   Fixed libpqxx headers

   Revision 1.9  2003/09/02 12:54:10  arjen
   Overloaded gnucomo_database::Field() to include the Result
   from a query as an argument.

   Revision 1.8  2003/08/17 11:39:33  arjen
   Changed the gnucomo_database class to the new PostgreSQL
   library, libpqxx

   Revision 1.7  2003/03/29 08:13:53  arjen
   New member function gnucomo_database::is_conected().

   Revision 1.6  2003/02/19 09:54:47  arjen
   Print the query on cerr, along with the error message when
   the query results in an error.

   Revision 1.5  2003/02/05 09:33:17  arjen
   gnucomo_database::new_notification() retruns the id number of the
   newly created notification record.

   Revision 1.4  2003/01/18 08:52:18  arjen
   New C++ function: gnucomo_database::new_notification()

   Revision 1.3  2002/11/09 08:04:27  arjen
   Added a reference to the GPL

   Revision 1.2  2002/11/04 10:13:36  arjen
   Use proper namespace for iostream classes

   Revision 1.1  2002/10/05 10:25:49  arjen
   Creation of gcm_input and a first approach to a web interface

*****************************/

/* static const char *RCSID = "$Id: database.h,v 1.12 2011-03-24 10:21:43 arjen Exp $"; */

#include <pqxx/connection>
#include <pqxx/transaction>
#include <pqxx/result>
#include "gnucomo_config.h"

/*
///////////////////////////////////////////////////////////////////////////
//  NAME           : gnucomo_database
//  BASECLASS      : 
//  MEMBERS        :
//  OPERATORS      :
//  METHODS        : is_connected - Return true if the database is connected
//
//  DESCRIPTION    : 
//
//  RELATIONS      :
//  SEE ALSO       :
//  LAST MODIFIED  : Aug 27, 2003
///////////////////////////////////////////////////////////////////////////
*/

class gnucomo_database
{
   gnucomo_config    *cfg;
   pqxx::connection  *dbconn;
   //pqxx::transaction<pqxx::serializable> *dbxact;
   pqxx::work *dbxact;

   pqxx::result      last_result;

public:

   gnucomo_database()
   {
      cfg = 0;
      dbconn = 0;
      dbxact = 0;
   }

   // Use the configuration to connect to the database

   gnucomo_database(gnucomo_config *c);

   //   A copy constructor and the assignement can copy
   //   the connection to the database must create a
   //   new Transaction object.

   gnucomo_database(const gnucomo_database &gdb);
   void operator = (const gnucomo_database &gdb);

   ~gnucomo_database();
   
   //  Error checking and handling.

   bool is_connected()
   {
      return dbconn != 0 && dbconn->is_open();
   }

   //  Low-level database access functions

   int Query(String qry)
   {
      try
      {
         last_result = dbxact->exec((char *)qry);
#ifdef DEBUG
      std::cerr << "Query " << qry << " returned "
                << last_result.size() << " tuples.\n";
#endif
      }
      catch (const pqxx::pqxx_exception &e)
      {
         std::cerr << "Error in QUERY " << qry << ":\n";
         std::cerr << e.base().what() << std::endl;
      }

      return last_result.size();
   }

   pqxx::result Result()
   {
      return last_result;
   }

   //  The field value of a specific result.

   String Field(pqxx::result res, int tuple, const char *fieldname)
   {
      return String(res[tuple][fieldname].c_str());
   }

   //  Use the result of the last query by default

   String Field(int tuple, const char *fieldname)
   {
      return String(last_result[tuple][fieldname].c_str());
   }

   //  Return the objectid of the host given its name.

   String find_host(const String hostname);

   //  Create a new notification.

   String new_notification(String objectid, String issue, String remark);
};

/*
///////////////////////////////////////////////////////////////////////////
//  NAME           : database_entity
//  BASECLASS      : 
//  MEMBERS        :
//  OPERATORS      :
//  METHODS        : 
//
//  DESCRIPTION    : 
//
//  RELATIONS      :
//  SEE ALSO       :
//  LAST MODIFIED  : Aug 19, 2005
///////////////////////////////////////////////////////////////////////////
*/

class database_entity
{
   const gnucomo_database  *db;
   String    table;

   bool fresh;       // Completely new, no tuple in the database
   bool changed;     // A database update is needed
   bool deleted;     // Tuple is to be deleted from the database

public:

  database_entity(const gnucomo_database &gdb, const String tbl)
  {
     db = &gdb;
     table = tbl;

     fresh = true;
     changed = false;
     deleted = false;
  }
};
